sealed abstract class Form
case class And(val left : Form, val right : Form) extends Form
case class Or(val left : Form, val right : Form) extends Form
case class Not(val body : Form) extends Form
case class Implies(val left : Form, val right : Form) extends Form
case class Var(name : String) extends Form

object Propositional {
  def eval(f : Form, e : Map[String, Boolean]) : Boolean = f match {
    case Var(name) => e(name)
    case Not(f) => !(eval(f,e))
    // add missing cases here
  }  

  // The nnf function body is just an example, you may wish to define
  // auxiliary functions
  def nnf(f : Form) : Form = f match {
    case Var(name) => f
    case Not(Var(name)) => f
    case Not(Not(f)) => nnf(f)
    // add missing cases here
  }

  def main(args : Array[String]) {
    val f1 = Not(Not(Var("p")))
    println("f1 = " + f1)

    val e = Map("p" -> true, "q" -> false)
    println("e = " + e)
    println("e(p) = " + e("p"))

    println("in interpretation e formula f1 as truth value:" + eval(f1,e))

    val f2 = Implies(Not(Or(Var("p"), Var("q"))),And(Not(Var("p")), Var("q")))
    println("f2 = " + f2)

    // println("The program will crash if eval is not finished:")
    // println("in interpretation e formula f2 as truth value:" + eval(f2,e))

    println("nnf(f1) = " + nnf(f1))

    // println("The program will crash if nnf is not finished:")
    // println("Negation-normal form of formula f2 is:" + nnf(f2))
  }
}
